from pathlib import Path
from collections import defaultdict, deque

from typing import Any, Type, Literal, Iterator, Optional


from weetags.utils import infer_dtype, infer_loader
from weetags.tree import Tree
from weetags.loaders import Loader, JlLoader, JsonLoader
from weetags.engine.engine import TreeEngine
from weetags.engine.schema import SimpleSqlField, SimpleSqlTable
from weetags.engine.schema import NodesTable, MetadataTable, IndexTable 

StrOrPath = str | Path
DataLoader = list[Type[Loader|JlLoader|JsonLoader]] | None
Data = list[dict[str, Any]] | list[Type[Loader|JlLoader|JsonLoader]] | list[StrOrPath] | None
Loaders = Literal["default", "lazy"]

class TreeBuilder(TreeEngine):
    data: DataLoader = None
    model: dict[str, Any] | None = None
    BATCH_SIZE = 500
    root_id = None

    def __init__(
        self,
        tree_name: str,
        database: Optional[str] = ":memory:",
        data: Optional[Data] = None,
        **params: Optional[Any]
        ) -> None:

        super().__init__(tree_name, database, **params)
        self._set_loaders(data)
        self._infer_model(None)
        self._collect_tables()

    @classmethod
    def build_tree(
        cls,
        tree_name: str,
        database: Optional[str] = ":memory:",
        data: Optional[Data] = None,
        indexes: Optional[list[str]] = None,
        read_only: Optional[bool] = False,
        replace: Optional[bool] = False,
        **params: Any
    ) -> Tree:
        builder = cls(tree_name, database, data, **params)
        if (builder.data is None and not builder._get_tables(tree_name)) or (replace and not builder.data):
            raise ValueError("You must initialize the TreeBuilder with a data or a builded database.")
        
        if replace:
            builder.drop_tree()
        if not builder._get_tables(tree_name) or replace:
            builder.build_tree_tables()
            if indexes:
                builder.build_indexes(indexes)
            builder.populate_tree()
        return Tree(tree_name=tree_name, database=database, read_only=read_only, **params)


    @property
    def iter_data(self) -> Iterator:
        for source in self.data:
            for node in source.loader():
                yield node

    def drop_tree(self) -> None:
        tables = self._get_tables(self.tree_name)
        for table in tables:
            self._drop(table[0])
    
    def build_tree_tables(self) -> None:
        tables = [self.tables["nodes"], self.tables["metadata"]]
        self._create_tables(*tables)

    def build_indexes(self, indexes: list[str]) -> None:
        nodes_table = self.tables["nodes"]
        for fname in indexes:
            field = getattr(self.tables["nodes"], fname.split(".")[0], None)
            if field is None:
                raise ValueError(f"Building Index: field {fname} does not exist")            

            if field.dtype == "JSON":
                target = fname.split(".")[0]
                path_name = fname.replace(".", "_")

                index_table = IndexTable.initialize(self.tree_name, path_name, "TEXT")
                self.tables[path_name] = index_table
                self._create_tables(index_table)
                # self._create_index(index_table, fname)
                self._create_triggers(index_table, target, fname)

            elif field.dtype == "JSONLIST":
                index_table = IndexTable.initialize(self.tree_name, fname, "TEXT")
                self.tables[fname] = index_table
                self._create_tables(index_table)
                self._create_index(index_table, fname)
                self._create_triggers(index_table, fname)
            else:
                self._create_index(nodes_table, fname)

    def populate_tree(self) -> None:
        if self.data is None:
            return
        
        batch, parent2children = deque(), defaultdict(list)
        for node in self.iter_data:
            if node.get("children", None) is None:
                node.update({"children":[]})

            # add directly the root node ... with meta data.
            if node["parent"] is None and self.root_id is None:
                self._build_root(node)
                continue

            # build batch
            batch.append(node)
            if node.get("parent", None):
                parent2children[node["parent"]].append(node["id"])

            # write db when batch size is attained
            if len(batch) == self.BATCH_SIZE:
                (batch, parent2children) = self._build_nodes(batch, parent2children)
                parent2children = self._add_remaining_children(parent2children)
                self.con.commit()
        # do the remaining nodes
        if len(batch) > 0:
            (batch, parent2children) = self._build_nodes(batch, parent2children)
            parent2children = self._add_remaining_children(parent2children)
            self.con.commit()

        # still need to setup metadata
        self._build_metadata()
        self.con.commit()

    def _build_root(self, node: dict[str, Any]) -> None:
        nodes_table = self.tables["nodes"]._name
        metadata_table = self.tables["metadata"]._name
        self.root_id = node["id"]
        self._builder_write_many(nodes_table, list(node.keys()), [list(node.values())])
        self._builder_write_many(metadata_table, ["nid", "depth", "is_root", "is_leaf"], [[node["id"], 0, True, False]])

    def _build_nodes(
        self,
        batch: deque[dict[str, Any]],
        parent2children: dict[str, list[str]]
    ) -> tuple[deque, dict[str, list[str]]]:

        nodes_table = self.tables["nodes"]._name
        k, v = list(batch[0].keys()), []
        while len(batch) > 0:
            node = batch.popleft()
            children = parent2children.pop(node["id"], [])
            node.update({"children": children})
            v.append(tuple(node.values()))
        self._builder_write_many(nodes_table, k, v, False)
        return (batch, parent2children)

    def _add_remaining_children(
        self,
        parent2children: dict[str, list[str]]
    ) -> dict[str, list[str]]:
        nodes_table = self.tables["nodes"]._name
        remains = self._get_children_from_ids(nodes_table, list(parent2children.keys()))
        for node in remains:
            new_children = parent2children.pop(node["id"])
            children = node["children"] + new_children
            self._builder_update(nodes_table, [("children", children)], [("id","=",node["id"])], commit=False)
        return parent2children

    def _build_metadata(self) -> None:
        nodes_table = self.tables["nodes"]._name
        metadata_table = self.tables["metadata"]._name
        root = self._get_children_from_id(nodes_table, self.root_id)
        current_layer, layers_size, queue, values = 1, defaultdict(int), deque(root["children"]), []
        layers_size[current_layer] += len(root["children"])
        while len(queue) > 0:
            nid = queue.popleft()
            children = self._get_children_from_id(nodes_table, nid)
            values.append(tuple((nid, current_layer, False, not any(children["children"]))))
            queue.extend(children["children"])

            layers_size[current_layer + 1] += len(children["children"])
            layers_size[current_layer] -= 1
            if layers_size[current_layer] == 0:
                current_layer += 1

            if len(values) == self.BATCH_SIZE:
                self._builder_write_many(metadata_table, ["nid", "depth", "is_root", "is_leaf"], values)
                values = []
        if len(values) > 0:
            self._builder_write_many(metadata_table, ["nid", "depth", "is_root", "is_leaf"], values)

    def _collect_tables(self) -> None:
        self.tables = {}
        tables = self._get_tables(self.tree_name)
        if len(tables) > 0:
            for table in tables:
                table_name = table[0]
                info = self._table_info(table_name)
                fk_info = self._table_fk_info(table_name)
                tkey = table_name.split("__")[1]    
                self.tables[tkey] = SimpleSqlTable.from_pragma(table_name, info, fk_info)
        elif self.model is None:
            raise ValueError("Input data files or a database into the TreeBuilder")
        else:
            nodes_fields = {k:SimpleSqlField(k,v) for k,v in self.model.items() if k not in ["nid", "id", "parent", "children"]}
            self.tables["nodes"] = NodesTable.initialize(self.tree_name, **nodes_fields)
            self.tables["metadata"] = MetadataTable.initialize(self.tree_name)

    def _set_loaders(self, data: Data, strategy: Loaders= "lazy") -> None:
        if data is None:
            return
        elif not isinstance(data, list):
            raise TypeError("TreeBuilder.data must be a list.")
        elif isinstance(data, list) and len(data) == 0:
            raise ValueError(f"TreeBuilder.data must contain at least one record or Filepath")
        
        self.data = []
        for d in data:
            if isinstance(d, dict):
                self.data.append(Loader(data))
                break
            elif isinstance(d, str):
                loader = infer_loader(d)
                self.data.append(loader(d, strategy))
            elif type(d) in [Loader, JlLoader, JsonLoader]:
                self.data.append(d)
            else:
                raise ValueError("data must be of type list[dict[str, Any]] | list[str] | None")
        
    def _infer_model(self, model: dict[str, Any] | None) -> None:
        if model is not None:
            return

        model = {}
        for record in self.iter_data:
            if "id" not in record.keys():
                raise KeyError("Data records must have an id field")
            if "parent" not in record.keys():
                raise KeyError("Data records must have a parent field.")

            for field, value in record.items():
                current_dtype = model.get(field, None)
                dtype = infer_dtype(value)
                if current_dtype is None:
                    model[field] = dtype
                elif dtype is None and current_dtype != "NULL":
                    continue
                elif current_dtype == "NULL" and dtype != "NULL":
                    model[field] = dtype
                elif current_dtype != dtype and dtype != "NULL":
                    raise ValueError(f"field `{field}` dtype is not consistent over the dataset.")
                    
        if model.get("id") != "TEXT":
            raise ValueError("Id must be a string")
        self.model = model


from __future__ import annotations

import json
import sqlite3
from sqlite3 import Cursor, Row
from sqlite3 import register_adapter, register_converter
from sqlite3 import PARSE_DECLTYPES

from typing import Any

import weetags.engine.sql as sql
from weetags.engine.sql import _SimpleSqlConverter, SqlConverter, OnConflict
from weetags.engine.schema import SimpleSqlTable, Namespace


Node = dict[str, Any]
Nodes = list[Node]
Conditions = list[list[tuple[str, str, Any] | str] | str]

class TreeEngine:
    tree_name: str
    database: str
    params: dict[str, Any]

    tables: dict[str, Any]
    namespaces: dict[str, Any]

    def __init__(self, tree_name: str, database: str = ":memory:", timeout: float = 5, **params) -> None:
        self.tree_name = tree_name
        self.database = database
        self.params = params
        self.timeout = timeout
        if database == ":memory":
            self.params.update({"cache":"shared"})

        self.con = sqlite3.connect(self.uri, detect_types=PARSE_DECLTYPES, uri=True, timeout=timeout)
        self.cursor = self.con.cursor()

        self.con.execute("PRAGMA foreign_keys=ON;")
        self.con.execute("PRAGMA case_sensitive_like=ON;")
        self.con.row_factory = self._record_factory
        register_adapter(list, self._serialize)
        register_adapter(dict, self._serialize)
        register_converter("JSON", self._deserialize) # type: ignore
        register_converter("JSONLIST", self._deserialize) # type: ignore

        self.tables = {}
        self.namespaces = {}

    @classmethod
    def from_pragma(cls, tree_name: str, database: str = ":memory:", timeout: float = 5, **params) -> TreeEngine:
        engine = cls(tree_name, database, **params)
        engine._build_tree_context(tree_name)
        return engine

    @property
    def uri(self) -> str:
        options = ""
        if bool(self.params):
            options = "?" + "&".join([f"{k}={v}" for k,v in self.params.items()])
        return f"file:{self.database}{options}"

    def _execute(self, query: str) -> None:
        self.cursor.execute(query)
        self.con.commit()

    def _execute_many(self, *queries: str) -> None:
        for query in queries:
            self.cursor.execute(query)
        self.con.commit()

    def _create_tables(self, *tables: SimpleSqlTable) -> None:
        queries = [table.create_table() for table in tables]
        self._execute_many(*queries)

    def _create_index(self, table: SimpleSqlTable, field_name: str) -> None:
        self._execute(table.create_index(field_name))

    def _create_json_extract_column(self, table: SimpleSqlTable, target_field: str, path: str) -> None:
        self._execute(table.create_json_extract_column(target_field, path))

    def _create_triggers(self, table: SimpleSqlTable, target_field: str,  path: str | None = None) -> None:
        nodes_table = self.tables["nodes"]
        self._execute(table.create_insert_trigger(target_field, nodes_table._name, path))
        self._execute(table.create_update_trigger(target_field, nodes_table._name, path))
        self._execute(table.create_delete_trigger(nodes_table._name))

    def _write_one(
        self,
        table_name: str,
        target_columns: list[str],
        values: list[Any],
        on_conflict: str = "none",
        commit: bool = True
    ) -> None:
        converter = SqlConverter(
            namespaces=self.namespaces,
            tables=self.tables,
            table_name=table_name,
            target_columns=target_columns,
            values=values,
            on_conflict=OnConflict(on_conflict),
        )
        stmt, values = converter.write_one()
        self.con.execute(stmt, values)
        if commit:
            self.con.commit()

    def _write_many(
        self,
        table_name: str,
        target_columns: list[str],
        values: list[list[Any]],
        on_conflict: str = "none",
        commit: bool = True
    ) -> None:
        converter = SqlConverter(
            namespaces=self.namespaces,
            tables=self.tables,
            table_name=table_name,
            target_columns=target_columns,
            values=values,
            on_conflict=OnConflict(on_conflict),
        )
        stmt, values = converter.write_many()
        self.con.execute(stmt, values)
        if commit:
            self.con.commit()

    def _builder_write_many(
        self,
        table_name: str,
        target_columns: list[str],
        values: list[list[Any]],
        commit: bool = True
        ) -> None:
        converter = _SimpleSqlConverter(
            table_name=table_name,
            target_columns=target_columns,
            values=values,
        )
        stmt, values = converter._write_many()
        self.con.executemany(stmt, values)
        if commit:
            self.con.commit()

    def _read_one(
        self,
        fields: list[str] | None = None,
        conditions: Conditions | None = None,
        order_by: list[str] | None = None,
        axis: int = 1
    ) -> Node:
        converter = SqlConverter(
            namespaces=self.namespaces,
            tables=self.tables,
            fields=fields,
            conds=conditions,
            order_by=order_by,
            axis=axis
        )
        stmt, values = converter.read_one()
        return self.con.execute(stmt, values).fetchone()

    def _read_many(
        self,
        fields: list[str] | None = None,
        conditions: Conditions | None = None,
        order_by: list[str] | None = None,
        limit: int | None = None,
        axis: int = 1,
    ) -> Nodes:
        converter = SqlConverter(
            namespaces=self.namespaces,
            tables=self.tables,
            fields=fields,
            conds=conditions,
            order_by=order_by,
            axis=axis,
            limit=limit
        )
        stmt, values = converter.read_many()
        return self.con.execute(stmt, values).fetchall()

    def _update(
        self,
        table_name: str,
        setter: list[tuple[str, Any]],
        conditions: Conditions | None = None,
        commit: bool = True
    ) -> None:
        converter = SqlConverter(
            namespaces=self.namespaces,
            tables=self.tables,
            table_name=table_name,
            conds=conditions,
            setter=setter
        )
        stmt, values = converter.update()
        self.con.execute(stmt, values)
        if commit:
            self.con.commit()


    def _builder_update(
        self,
        table_name: str,
        setter: list[tuple[str, Any]],
        conditions: list[tuple[str, str, Any]] | None = None,
        commit: bool = True    
        ) -> None:
        """primitive update builder. don't use SqlConverter."""
        converter = _SimpleSqlConverter(
            table_name=table_name,
            conds=conditions,
            setter=setter
        )
        stmt, values = converter._update()
        self.con.execute(stmt, values)
        if commit:
            self.con.commit()

    def _delete(self, conditions: Conditions | None = None, commit: bool = True) -> None:
        converter = SqlConverter(
            namespaces=self.namespaces,
            tables=self.tables,
            conds=conditions
        )
        stmt, values = converter.delete()
        self.con.execute(stmt, values)
        if commit:
            self.con.commit()

    def _drop(self, table_name: str) -> None:
        query = sql.DROP.format(table_name=table_name)
        self.cursor.execute(query)
        self.con.commit()

    def _table_info(self, table_name: str) -> list[tuple]:
        query = sql.INFO.format(table_name=table_name)
        return self.cursor.execute(query).fetchall()

    def _table_fk_info(self, table_name: str) -> list[tuple]:
        query = sql.FK_INFO.format(table_name=table_name)
        return self.cursor.execute(query).fetchall()

    def _table_size(self, table_name: str) -> int:
        query = sql.TABLE_SIZE.format(table_name=table_name)
        return self.cursor.execute(query).fetchone()[0]

    def _max_depth(self, table_name: str) -> int:
        query = sql.TREE_DEPTH.format(table_name=table_name)
        return self.cursor.execute(query).fetchone()[0]

    def _get_tables(self, tree_name: str) -> list[str]:
        query = sql.TABLE_NAMES.format(tree_name=tree_name)
        return self.cursor.execute(query).fetchall()

    def _get_children_from_ids(self, nodes_table: str, ids: list[str]) -> list[tuple[str, list[str]]]:
        converter = _SimpleSqlConverter(table_name=nodes_table, values=ids)
        query, values = converter._children_from_ids()
        return self.con.execute(query, values).fetchall()

    def _get_children_from_id(self, nodes_table: str, id: str):
        converter = _SimpleSqlConverter(table_name=nodes_table, values=[id])
        query, values = converter._children_from_id()
        return self.con.execute(query, values).fetchone()

    def _get_user(self, username: str) -> dict[str, Any] | None:
        return self.con.execute(sql.GET_USER, [username]).fetchone()

    def _get_restriction(self, tree_name: str, blueprint: str) -> dict[str, Any] | None:
        return self.con.execute(sql.GET_RESTRICTION, [tree_name, blueprint]).fetchone()

    def _add_users(self, *users) -> None:
        for settings in users:
            converter = _SimpleSqlConverter(
                table_name="weetags__users", 
                target_columns=settings.fields, 
                values=[settings.values]
            )
            stmt, values = converter._write_many()
            self.con.executemany(stmt, values)
        self.con.commit()

    def _add_restrictions(self, *restrictions) -> None:
        for settings in restrictions:
            converter = _SimpleSqlConverter(
                table_name="weetags__restrictions", 
                target_columns=settings.fields, 
                values=[settings.values]
            )
            stmt, values = converter._write_many()
            self.con.executemany(stmt, values)
        self.con.commit()

    @staticmethod
    def _serialize(data: dict[str, Any] | list[Any]) -> str:
        return json.dumps(data)

    @staticmethod
    def _deserialize(data: str) -> dict[str, Any] | list[Any]:
        return json.loads(data)

    @staticmethod
    def _record_factory(cursor: Cursor, row: Row) -> dict[str, Any]:
        fields = [column[0] for column in cursor.description]
        return {k:v for k,v in zip(fields, row)}
    
    @staticmethod
    def condition_anchor(op: str, values: Any) -> str:
        """define the right anchor for the given condition operator."""
        anchor = "?"
        if op.lower() == "in" and isinstance(values, list):
            anchors = ' ,'.join(["?" for _ in range(len(values))])
            anchor = f"({anchors})"
        return anchor

    def _build_tree_context(self, tree_name: str) -> None:
        """Build tables and namespaces collections from db pragma"""
        self.tables = {}
        self.namespaces = {}

        tables = self._get_tables(tree_name)
        if len(tables) == 0:
            raise ValueError(
                f"tree: {tree_name} is currently not builded.",
                "Consider building the tree first with the TreeBuilder"
            )

        for table in tables:
            table_name = table[0]
            info = self._table_info(table_name)
            fk_info = self._table_fk_info(table_name)
            table_type = table_name.split("__")[1]

            table_repr = SimpleSqlTable.from_pragma(table_name, info, fk_info)
            for fname, f in table_repr.iter_fields:
                current_namespace = self.namespaces.get(fname, None)
                if table_type not in ["metadata", "nodes"] and fname in ("nid", "elm_idx"):
                    continue
                elif current_namespace is None:
                    self.namespaces[fname] = Namespace(
                        table = table_repr._name,
                        index = table_repr._name,
                        fname = fname,
                        ftype = f.dtype
                    )
                else:
                    current_namespace.index_table = table_repr._name
            self.tables[table_type] = table_repr



from __future__ import annotations

import random
from abc import ABC
from hashlib import sha256
from pathlib import Path
from string import ascii_letters, digits
from attrs import define, field, validators


from typing import Any, Iterator

import weetags.engine.sql as sql

StrOrPath = str | Path
Users = Restricions = list[dict[str, Any]] | None

CHARS = ascii_letters + digits

@define(slots=True, kw_only=True)
class User:
    username: str = field(validator=[validators.instance_of(str)])
    password: str = field(validator=[validators.instance_of(str)])
    password_sha256: str = field(default=None)
    auth_level: list[str] = field()
    salt: str = field(default=None)
    max_age: int = field(validator=[validators.instance_of(int)])

    def __attrs_post_init__(self):
        if self.password_sha256 is None and self.salt is None:
            self.generate_salt()
            self.generate_sha256()

    def generate_salt(self, size: int = 16):
        self.salt = "".join([random.choice(ascii_letters) for _ in range(size)])

    def generate_sha256(self):
        salted_password = self.salt + self.password
        self.password_sha256 = sha256(salted_password.encode()).hexdigest()

    @property
    def fields(self) -> list[str]:
        return ["username", "password_sha256", "auth_level", "salt", "max_age"]

    @property
    def values(self) -> list[Any]:
        return [self.username, self.password_sha256, self.auth_level, self.salt, self.max_age]

@define(slots=True, kw_only=True)
class Restriction:
    tree: str = field(validator=[validators.instance_of(str)])
    blueprint: str = field(validator=[validators.instance_of(str)])
    auth_level: list[str] = field(validator=[validators.instance_of(list)])

    def create_restriction(self) -> tuple[str, list[Any]]:
        ...

    @property
    def fields(self) -> list[str]:
        return ["tree", "blueprint", "auth_level"]

    @property
    def values(self) -> list[Any]:
        return [self.tree, self.blueprint, self.auth_level]


@define(kw_only=True, slots=False)
class Namespace:
    table: str = field(validator=[validators.instance_of(str)])
    index: str = field(validator=[validators.instance_of(str)])
    fname: str = field(validator=[validators.instance_of(str)])
    ftype: str = field(validator=[validators.instance_of(str)])

    def is_joinable(self) -> bool:
        return self.index.split("__")[1] != "nodes"

    def is_metadata(self) -> bool:
        return self.table.split("__")[1] == "metadata"

    def select(self) -> str:
        return f"{self.table}.{self.fname}"

    def join(self, to_table: str) -> str:
        if self.index.endswith("nodes"):
            raise KeyError("__nodes is not joinable")
        return f"JOIN {self.index} ON {to_table}.id = {self.index}.nid"

    def where(self, op: str, values: Any) -> tuple[str, Any]:
        fname, op, values = self._prepare(op, values)
        return ((f"{fname} {op} {self._set_anchor(op, values)}", values))

    def _prepare(self, op: str, values: Any) -> tuple[str, Any]:
        if op.lower() == "ilike" and isinstance(values, list):
            v = []
            for val in values:
                if not isinstance(val, str):
                    raise ValueError(f"ILIKE operator must compare Strings. `{values}` is not a string")
                v.append(val.upper())
            return (f"UPPER({self.index}.{self.fname})", "LIKE", v)
        elif op.lower() == "ilike" and isinstance(values, str):
            return (f"UPPER({self.index}.{self.fname})", "LIKE", values.upper())
        elif op.lower() == "ilike":
            raise ValueError(f"ILIKE operator must compare Strings. `{values}` is not a string")
        else:
            return (f"{self.index}.{self.fname}", op, values)

    def _set_anchor(self, op: str, values: Any) -> str:
        if op.lower() == "in" and isinstance(values, list):
            return f"({' ,'.join(['?' for _ in range(len(values))])})"
        elif op.lower() == "in" and not isinstance(values, list):
            return "(?)"
        else:
            return "?"

@define(slots=False)
class SimpleSqlField:
    name: str = field()
    dtype: str = field()
    pk: bool = field(default=False)
    fk: str | None = field(default=None)
    nullable: bool = field(default=True)
    unique: bool = field(default=False)
    serial: bool = field(default=False)

    @property
    def foreign_key(self) -> str:
        if self.fk is None:
            raise ValueError("This field has no FK bound")
        table_name, field_name = self.fk.split('.')
        return f"FOREIGN KEY ({self.name}) REFERENCES {table_name}({field_name}) ON DELETE CASCADE"

    def to_sql(self) -> str:
        return f"{self.name} {self.dtype} {self._options()}"

    def _options(self) -> str:
        sql = {
            ("nullable", False): "NOT NULL",
            ("unique", True): "UNIQUE",
            ("serial", True): "AUTOINCREMENT"
        }
        options = []
        for k,v in sql.items():
            field_value = getattr(self, k[0], None)
            sql_value = sql.get((k[0], field_value), None) # type: ignore
            if sql_value is not None:
                options.append(v)
        return " ".join(options)




@define(slots=False)
class SimpleSqlTable(ABC):
    _name: str = field()

    @classmethod
    def from_pragma(cls, _name: str, table_info: list[tuple], fk_info: list[tuple]) -> SimpleSqlTable:
        table = cls(_name)
        fks = {f[3]:f"{f[2]}.{f[4]}" for f in fk_info}
        for f in table_info: # iterate over sql fields
            simple_field = SimpleSqlField(f[1], f[2],pk=bool(f[5]),fk=fks.get(f[1], None))
            setattr(table, f[1], simple_field)
        return table

    @property
    def fields(self) -> dict[str, SimpleSqlField]:
        return {k:attr for k,attr in self.__dict__.items() if isinstance(attr, SimpleSqlField)}

    @property
    def iter_fields(self) -> Iterator[tuple[str, SimpleSqlField]]:
        for k,attr in self.__dict__.items():
            if isinstance(attr, SimpleSqlField):
                yield (k, attr)

    def create_table(self) -> str:
        fields, pk, fk = [], [], []
        for _, f in self.iter_fields:
            fields.append(f.to_sql())
            if f.fk is not None:
                fk.append(f.foreign_key)
            if f.pk is not False:
                pk.append(f.name)
        data = ", ".join(fields + [sql.pk_to_sql(pk)] + fk)
        return sql.CREATE_TABLE.format(table_name=self._name, fields=data)

    def create_index(self, field_name: str) -> str:
        return sql.CREATE_INDEX.format(table_name=self._name, field_name=field_name)

    def create_json_extract_column(self, target_field: str, path: str) -> str:
        return sql.CREATE_EXTRACT_COLUMN.format(table_name=self._name, target_field=target_field, path=path)

    def create_insert_trigger(self, target_field: str, target_table: str | None = None, path: str | None = None) -> str:
        if path is not None:
            base = path.split(".")[0]
            inner_path = ".".join(path.split(".")[1:])
            fname = path.replace(".", "_")
            trigger = sql.ADD_JSON_TRIGGER.format(table_name=self._name, target_table=target_table, target_field=fname, base=base, path=inner_path)
        elif path is None and target_table:
            trigger = sql.ADD_JSONLIST_TRIGGER.format(table_name=self._name, target_table=target_table, target_field=target_field)
        else:
            raise ValueError("Cannot create insert trigger.")
        return sql.CREATE_TRIGGER.format(table_name=self._name, target_table=target_table, trigger=trigger)

    def create_delete_trigger(self, target_table: str) -> str:
        return sql.DELETE_TRIGGER.format(table_name=self._name, target_table=target_table)

    def create_update_trigger(self, target_field: str, target_table: str | None = None, path: str | None = None) -> str:
        if path is not None:
            base = path.split(".")[0]
            inner_path = ".".join(path.split(".")[1:])
            fname = path.replace(".", "_")
            trigger = sql.UPDATE_JSON_TRIGGER.format(table_name=self._name, target_table=target_table, target_field=fname, base=base, path=inner_path)
        elif path is None and target_table:
            trigger = sql.UPDATE_JSONLIST_TRIGGER.format(table_name=self._name, target_table=target_table, target_field=target_field)
        else:
            raise ValueError("Cannot Create update trigger")
        return sql.UPDATE_TRIGGER.format(table_name=self._name, target_table=target_table, target_field=target_field, trigger=trigger)


@define(slots=False)
class NodesTable(SimpleSqlTable):
    _name: str = field()
    id: SimpleSqlField = field(default=SimpleSqlField("id", "TEXT", pk=True, nullable=False, unique=True))
    parent: SimpleSqlField = field(default=SimpleSqlField("parent", "TEXT"))
    children: SimpleSqlField = field(default=SimpleSqlField("children", "JSONLIST", nullable=False))

    @classmethod
    def initialize(cls, _name: str, **fields: SimpleSqlField) -> NodesTable:
        table_name = f"{_name}__nodes"
        table = cls(table_name)
        for k,v in fields.items():
            if k in table.__dict__.keys():
                raise ValueError("Table Field already exist")
            setattr(table, k, cls.validate_field(v))
        return table

    @classmethod
    def validate_field(cls, v: SimpleSqlField) -> SimpleSqlField:
        if isinstance(v, SimpleSqlField) is False:
            raise ValueError("Field must be of type: SimpleSqlField")
        return v

@define(slots=False)
class MetadataTable(SimpleSqlTable):
    _name: str = field()
    nid: SimpleSqlField = field(default=SimpleSqlField("nid", "TEXT", pk=True, unique=True))
    depth: SimpleSqlField = field(default=SimpleSqlField("depth", "INTEGER", nullable=False))
    is_root: SimpleSqlField = field(default=SimpleSqlField("is_root", "BOOL", nullable=False))
    is_leaf: SimpleSqlField = field(default=SimpleSqlField("is_leaf", "BOOL", nullable=False))

    @classmethod
    def initialize(cls, _name: str) -> MetadataTable:
        nodes_table = f"{_name}__nodes"
        return cls(
            f"{_name}__metadata",
            nid=SimpleSqlField("nid", "TEXT", pk=True, fk=f"{nodes_table}.id")
        )

@define(slots=False)
class IndexTable(SimpleSqlTable):
    _name: str = field()
    nid: SimpleSqlField = field()
    value: SimpleSqlField = field()
    elm_idx: SimpleSqlField = field(default=SimpleSqlField("elm_idx", "INTEGER", pk=True, nullable=False))

    @classmethod
    def initialize(cls, _name: str, field_name: str, field_dtype: str) -> IndexTable:
        table_name=f"{_name}__{field_name}"
        nodes_table = f"{_name}__nodes"
        table = cls(
            table_name,
            nid= SimpleSqlField("nid", "TEXT", fk=f"{nodes_table}.id", pk=True, nullable=False),
            value= SimpleSqlField(field_name, field_dtype, pk=True)
        )
        return table

@define(slots=False)
class UsersTable(SimpleSqlTable):
    _name: str = field(default="weetags__users")
    username: SimpleSqlField = field(default=SimpleSqlField("username", "TEXT", pk=True, nullable=False))
    password_sha256: SimpleSqlField = field(default=SimpleSqlField("password_sha256", "TEXT", nullable=False))
    auth_level: SimpleSqlField = field(default=SimpleSqlField("auth_level", "JSONLIST", nullable=False))
    salt: SimpleSqlField = field(default=SimpleSqlField("salt", "TEXT", nullable=False))
    max_age: SimpleSqlField = field(default=SimpleSqlField("max_age", "INTEGER", nullable=False))

@define(slots=False)
class RestrictionsTable(SimpleSqlTable):
    _name: str = field(default="weetags__restrictions")
    tree: SimpleSqlField = field(default=SimpleSqlField("tree", "TEXT", pk=True, nullable=False))
    blueprint: SimpleSqlField = field(default=SimpleSqlField("blueprint", "TEXT", pk=True, nullable=False))
    auth_level: SimpleSqlField = field(default=SimpleSqlField("auth_level", "JSONLIST", nullable=False))



import sqlite3
import traceback
from pathlib import Path

from typing import Any, Callable

from weetags.engine.sql import DTYPES
from weetags.loaders import JlLoader, JsonLoader

StrOrPath = str | Path




def infer_loader(path: StrOrPath) -> JlLoader | JsonLoader:
    ext = path.split(".")[-1]
    loaders = {
        "json": JsonLoader,
        "jl": JlLoader,
        "jsonlines": JlLoader
    }
    loader = loaders.get(ext, None)
    if loader is None:
        raise ValueError("non recognized file type")
    return loader

def infer_dtype(value: Any) -> str:
    if isinstance(value, bool):
        dtype = "BOOL"
    elif isinstance(value, int):
        dtype = "INTEGER"
    elif isinstance(value, str):
        dtype = "TEXT"
    elif value is None:
        dtype = "NULL"
    elif isinstance(value, float):
        dtype = "REAL"
    elif isinstance(value, list):
        dtype = "JSONLIST"
    else:
        dtype = "JSON"
    return dtype

def valid_creation(f: Callable):
    def wrapped(*args):
        tree, node = args
        node_table = tree.tables.get("nodes")
        for _, field in node_table.iter_fields:
            value = node.get(field.name, None)
            if field.dtype == "JSON" and value is None:
                node.update({field.name:{}})
            elif field.dtype == "JSONLIST" and value is None:
                node.update({field.name:[]})
            if value is not None and isinstance(value, DTYPES[field.dtype]) is False:
                raise ValueError(f"node field {field.name} either doesn't exist or has wrong dtype.")

        parent = node.get("parent", False)
        if parent is False: # none  is for root
            raise ValueError(f"A node must have a `parent` field")
        return f(tree, node)
    return wrapped

def valid_update(f: Callable):
    def wrapped(*args):
        tree, cond, set_values = args
        node_table = tree.tables.get("nodes")
        meta_table = tree.tables.get("metadata")
        for k,v in set_values:
            field = getattr(node_table, k, None) or getattr(meta_table, k, None)
            if field.name in ["id","nid","parent","children", "depth", "is_root", "is_leaf"]:
                raise KeyError(f"You cannot update the following fields: [`id`, `nid`, `parent`, `children`, `depth`, `is_root`, `is_leaf`]")
            if field is None or isinstance(v, DTYPES[field.dtype]) is False:
                raise ValueError(f"node field {k} either doesn't exist or has wrong dtype.")
        return f(tree, cond, set_values)
    return wrapped

def valid_append(f: Callable):
    def wrapped(*args):
        tree, cond, fname, values = args
        node_table = tree.tables.get("nodes")
        meta_table = tree.tables.get("metadata")
        field = getattr(node_table, fname, None) or getattr(meta_table, fname, None)
        if fname in ["id","nid","parent","children", "depth", "is_root", "is_leaf"]:
            raise KeyError(f"You cannot update the following fields: [`id`, `nid`, `parent`, `children`, `depth`, `is_root`, `is_leaf`]")
        if field is None or field.dtype not in ["JSON","JSONLIST"]:
            raise TypeError("field_name must reference field containing a collection such as a list or a dict")
        return f(tree, cond, fname, values)
    return wrapped    


def apply_handler(decorator: Callable):
    def decorate(cls):
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)):
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls
    return decorate

def ErrorHandler(f: Callable):
    def wrapped(*args, **kwargs):
        cls = args[0]
        ok, max_tries, tries = False, 5, 0
        while ok is False and tries < max_tries:
            try:
                res = f(*args, **kwargs)
                ok = True
                return res
            except sqlite3.IntegrityError as e:
                tries += 1
                print(f"INTEGRITY {e}")

            except sqlite3.DatabaseError as e:
                print(f"ERROR: {e}")
                tries += 1
            except Exception:
                tries += 1
                traceback.print_exc(limit=0)
                exit(1)
        
    return wrapped



import json
from typing import Any, Generator

from weetags.exceptions import NotImplemented


Payload = dict[str, Any]
TableName = FieldName = str

class Loader(object):
    def __init__(self, data: list[Payload]) -> None:
        self.data = data
        self.loader = self.default_loader

    def default_loader(self) -> Generator:
        for line in iter(self.data):
            yield line


class JsonLoader(Loader):
    def __init__(self, fp: str, strategy: str= "default") -> None:
        self.fp = fp
        self.loader = {
            "default": self.default_loader,
            "lazy": self.lazy_loader
        }[strategy]

    def default_loader(self) -> Generator:
        with open(self.fp) as f:
            data = iter(json.load(f))
            while line := next(data, None):
                yield line

    def lazy_loader(self) -> Generator:
        raise NotImplemented()


class JlLoader(Loader):
    def __init__(self, fp: str, strategy: str= "default") -> None:
        self.fp = fp
        self.loader = {
            "default": self.default_loader,
            "lazy": self.lazy_loader
        }[strategy]

    def default_loader(self) -> Generator:
        with open(self.fp) as f:
            data = iter([json.loads(line) for line in f.readlines()])
            while line := next(data, None):
                yield line

    def lazy_loader(self) -> Generator:
        with open(self.fp) as f:
            while line := f.readline():
                yield json.loads(line.strip("\n"))

    def __call__(self) -> Any:
        yield from self.loader()



from attrs import Attribute

from typing import Any

class WeetagsException(Exception):
    pass

class NotImplemented(WeetagsException):
    message = """Not implemented"""
    status = 501
    def __init__(self, ) -> None:
        super().__init__(self.message)
        
class ServiceUnavailable(WeetagsException):
    message = """Service Unavailable"""
    status = 503
    def __init__(self, ) -> None:
        super().__init__(self.message)
        
class TreeDoesNotExist(WeetagsException):
    message = """the tree "{name}" is unknown (existing trees: [{trees}])"""
    status = 400
    def __init__(self, name: str, trees: list[str]) -> None:
        trees = ", ".join(trees)
        super().__init__(self.message.format(name=name, trees=trees))

class ParsingError(WeetagsException):
    message = """Parameter(name: {name}, value: {value}) must be of type: {dtype}"""
    status = 400
    def __init__(self, attribute: Attribute, value: Any, dtype_annot: str) -> None:
        super().__init__(self.message.format(name=attribute.name, value=value, dtype=dtype_annot))

class CoversionError(WeetagsException):
    message = """Parameter(value: {value}) Cannot be converted into type: {dtype}"""
    status = 400
    def __init__(self, value: Any, dtype_annot: str) -> None:
        super().__init__(self.message.format(value=value, dtype=dtype_annot))

class UnknownRelation(WeetagsException):
    message = """Relation "{relation}" is unknown. possible relations : [{relations}]"""
    status = 400
    def __init__(self, relation: str, relations: list[str]) -> None:
        relations = ", ".join(relations)
        super().__init__(self.message.format(relation=relation, relations=relations))

class OutputError(WeetagsException):
    message = """{relation} does not return the expected type ({expected})"""
    status = 400
    def __init__(self, relation: str, expected:str) -> None:
        super().__init__(self.message.format(relation=relation, expected=expected))


class MissingLogin(WeetagsException):
    message = """You must provide a username and a password"""
    status = 401
    def __init__(self, ) -> None:
        super().__init__(self.message)

class InvalidLogin(WeetagsException):
    message = """Invalid username or password"""
    status = 401
    def __init__(self, ) -> None:
        super().__init__(self.message)

class InvalidToken(WeetagsException):
    message = """Invalid authorization token"""
    status = 401
    def __init__(self, ) -> None:
        super().__init__(self.message)
        
class AccessDenied(WeetagsException):
    message = """Access denied"""
    status = 401
    def __init__(self, ) -> None:
        super().__init__(self.message)

class AuthorizationTokenRequired(WeetagsException):
    message = """This endpoint is protected. Get an authorization token from `/login` endpoint."""
    status = 401
    def __init__(self, ) -> None:
        super().__init__(self.message)

class OutatedAuthorizationToken(WeetagsException):
    message = """Authorization token is outdated."""
    status = 401
    def __init__(self, ) -> None:
        super().__init__(self.message)




from attrs import define, field, validators
from collections.abc import Sequence
from enum import StrEnum

from typing import Any

from weetags.engine.schema import SimpleSqlTable, Namespace

Node = dict[str, Any]
Nodes = list[Node]
Conditions = list[list[tuple[str, str, Any] | str] | str]

DTYPES = {"TEXT": str, "INTEGER": int, "JSONLIST": list, "JSON": dict, "BOOL": bool}

# CREATE
CREATE_TABLE = "CREATE TABLE  IF NOT EXISTS {table_name} ({fields});"
CREATE_INDEX = "CREATE INDEX IF NOT EXISTS idx_{table_name}_{field_name} ON {table_name}({field_name});"
CREATE_EXTRACT_COLUMN = "ALTER TABLE {table_name} ADD COLUMN {target_field}_{path} TEXT AS (json_extract({target_field}, '{path}'))"

# TRIGGERS
ADD_JSON_TRIGGER = """\
INSERT INTO {table_name}({target_field}, nid, elm_idx)
SELECT json_extract({base}, "$.{path}"), {target_table}.id, 1 FROM {target_table} WHERE {target_table}.id = NEW.id;
"""

UPDATE_JSON_TRIGGER = """\
DELETE FROM {table_name} WHERE nid = OLD.id;
INSERT INTO {table_name}({target_field}, nid, elm_idx)
SELECT json_extract({base}, "$.{path}"), {target_table}.id, 1 FROM {target_table} WHERE {target_table}.id = NEW.id;
"""

ADD_JSONLIST_TRIGGER = """\
INSERT INTO {table_name}({target_field}, nid, elm_idx)
SELECT j.value, {target_table}.id, j.key FROM {target_table}, json_each(NEW.{target_field}) as j WHERE {target_table}.id = NEW.id;
"""

UPDATE_JSONLIST_TRIGGER = """\
DELETE FROM {table_name} WHERE nid = OLD.id;
INSERT INTO {table_name}({target_field}, nid, elm_idx)
SELECT j.value, {target_table}.id, j.key FROM {target_table}, json_each(NEW.{target_field}) as j WHERE {target_table}.id = NEW.id;
"""

CREATE_TRIGGER = """\
CREATE TRIGGER {table_name}__insert_trigger AFTER INSERT ON {target_table} BEGIN
{trigger}
END;
"""
DELETE_TRIGGER = """\
CREATE TRIGGER {table_name}__delete_trigger AFTER DELETE ON {target_table} BEGIN
DELETE FROM {table_name} where nid = OLD.id;
END;
"""
UPDATE_TRIGGER = """\
CREATE TRIGGER {table_name}__update_on_{target_field}_trigger AFTER UPDATE OF {target_field} ON {target_table} BEGIN
{trigger}
END;
"""


## infos
INFO = "PRAGMA table_info({table_name});"
FK_INFO = "PRAGMA foreign_key_list({table_name});"
TABLE_SIZE = "SELECT COUNT(*) FROM {table_name};"
TREE_DEPTH = "SELECT MAX(depth) FROM {table_name};"
TABLE_NAMES = "SELECT name FROM sqlite_master WHERE type='table' AND name LIKE '{tree_name}__%';"

# actions
DROP = "DROP TABLE IF EXISTS {table_name};"
WRITE = "INSERT {on_conflict} INTO {table_name}({col_names}) VALUES({anchors});"
READ_ONE = "SELECT {fields} FROM {node_table} {joins} {conditions} {order} {axis};"
READ_MANY = "SELECT {fields} FROM {node_table} {joins} {conditions} {order} {axis} {limit};"
UPDATE = "UPDATE {table_name} SET {setter} {conditions};"
DELETE = "DELETE FROM {node_table} {conditions};"

SEARCH_SUBQUERY = "id IN ({subquery})"
CHILDREN_FROM_IDS = "SELECT id, children FROM {table_name} WHERE id IN ({anchors});"
CHILDREN_FROM_ID = "SELECT id, children FROM {table_name} WHERE id=?;"
GET_USER = "SELECT username, password_sha256, auth_level, salt, max_age FROM weetags__users WHERE username=?"
GET_RESTRICTION = "SELECT auth_level FROM weetags__restrictions WHERE tree=? AND blueprint=?"

def pk_to_sql(pk: list[str]) -> str:
    s = ""
    if pk:
        s = f"PRIMARY KEY ({', '.join(pk)})"
    return s

class SqlOperator(StrEnum):
    Eq = "="
    Diff = "!="
    Glob = "GLOB"
    Gt = ">"
    Ge = ">="
    Ngt = "!>"
    Lt = "<"
    Le = "<="
    Nlt = "!<"
    Like = "LIKE"
    Ilike = "ILIKE"
    In = "IN"
    Not_in = "NOT IN"
    Is = "IS"
    Is_not = "IS NOT"

    def values():
        return [s.value for s in SqlOperator]

class OnConflict(StrEnum):
    Update = "update"
    Ignore = "ignore"
    Rollback = "rollback"
    Abort = "abort"
    Fail = "fail"
    Replace = "replace"
    NONE = "none"

    def sql(self) -> str:
        if str(self) == "none":
            return ""
        return f"OR {str(self).upper()}"


def intOrNone(instance, attribute, value):
    if value is None:
        return
    if not isinstance(value, int):
        raise TypeError(f"attribute {attribute} must be of type int.")

def listOrNone(instance, attribute, value):
    if value is None:
        return
    if not isinstance(value, list):
        raise TypeError(f"attribute {attribute} must be of type list.")

def onConflictOrNone(instance, attribute, value):
    if value is None:
        return
    if not isinstance(value, OnConflict):
        raise TypeError(f"attribute {attribute} must be of type OnConflict.")


@define(kw_only=True)
class _SimpleSqlConverter:
    """
    Works without any namespace nor tables structure.
    Ideally used during building time by the treeBuilder
    """
    table_name: str | None = field(default=None)
    target_columns: list[str] | None = field(default=None, validator=[listOrNone])
    values: list[Any]|list[list[Any]]|None = field(default=None, validator=[listOrNone])
    setter: list[tuple[str, Any]] | None = field(default=None, validator=[listOrNone])
    conds: list[tuple[str, str, Any]] | None = field(default=None, validator=[listOrNone])


    def _children_from_ids(self) -> tuple[str, list[Any]]:
        anchors = self.anchors(self.values)
        stmt = CHILDREN_FROM_IDS.format(table_name=self.table_name, anchors=anchors)
        return (stmt, self.values)

    def _children_from_id(self) -> tuple[str, list[Any]]:
        stmt = CHILDREN_FROM_ID.format(table_name=self.table_name)
        return (stmt, self.values)

    def _write_many(self) -> tuple[str, list[list[Any]]]:
        # add validation for columns names ? and table_name ?

        if not all([self.table_name, self.target_columns, self.values]):
            raise ValueError()

        columns = ", ".join(self.target_columns) # type: ignore
        anchors = self.anchors(self.values[0]) # type: ignore
        stmt =  WRITE.format(table_name=self.table_name, col_names=columns, anchors=anchors, on_conflict="")
        return (stmt, self.values) # type: ignore

    def _update(self) -> tuple[str, list[Any]]:
        def parse_simple_conditions(conds: list[tuple[str, str, Any]] | None) -> tuple[str, list[Any]]:
            if conds is None:
                return ("", [])

            values, conditions = [], []
            for field, op, val in conds:
                anchors = f"({self.condition_anchor(op, val)})"
                conditions.append(f"{field} {op} {anchors}")
                values.append(val)
            return ("WHERE " + ' AND '.join(conditions), values)

        def parse_simple_setter(setter: list[tuple[str, Any]]) -> tuple[str, list[Any]]:
            values, fields = [], []
            [(fields.append(f"{field} = ?"), values.append(val)) for field, val in setter]
            return (", ".join(fields), values)

        if self.setter is None:
            raise ValueError(f"You must Set some pairs of key values to update.")

        conditions, cvalues = parse_simple_conditions(self.conds)
        setter, svalues = parse_simple_setter(self.setter)
        stmt = UPDATE.format(table_name=self.table_name, setter=setter, conditions=conditions)
        return (stmt, svalues + cvalues)


    @staticmethod
    def condition_anchor(op: str, values: Any) -> str:
        """define the right anchor for the given condition operator."""
        anchor = "?"
        if op.lower() == "in" and isinstance(values, list):
            anchors = ", ".join(["?" for _ in range(len(values))])
            anchor = f"({anchors})"
        return anchor

    @staticmethod
    def anchors(values: list[Any]) -> str:
        return ' ,'.join(["?" for _ in range(len(values))])

    @staticmethod
    def update_setter(set: list[tuple[str, Any]]) -> tuple[str, Any]:
        values, fields = [], []
        [(fields.append(f"{field} = ?"), values.append(val)) for field, val in set]
        return (", ".join(fields), values)


@define(kw_only=True)
class SqlConverter:
    namespaces: dict[str, Namespace] = field()
    tables: dict[str, SimpleSqlTable] = field()

    table_name: str | None = field(default=None)
    target_columns: list[str] | None = field(default=None, validator=[listOrNone])
    values: list[Any]|list[list[Any]]|None = field(default=None, validator=[listOrNone])
    on_conflict: OnConflict|None = field(default=None, validator=[onConflictOrNone])
    conds: list[list[tuple[str, str, Any] | str] | str] | None = field(default=None, validator=[listOrNone])
    limit: int | None = field(default=None, validator=[intOrNone])
    fields: list[str] | None = field(default=None, validator=[listOrNone])
    order_by: list[str] | None = field(default=None, validator=[listOrNone])
    axis: int = field(default=1, converter=int, validator=[validators.instance_of(int)])
    setter: list[tuple[str, Any]] | None = field(default=None, validator=[listOrNone])

    def write_one(self) -> tuple[str, list[Any]]:

        # add validation for columns names ? and table_name ?

        if not all([self.table_name, self.target_columns, self.values]):
            raise ValueError()
        target_table = self.tables.get(self.table_name, None) # type: ignore
        if target_table is None:
            raise KeyError(f"Unknown table type: {self.table_name}")
        ttable_name = target_table._name
        columns = " ,".join(self.target_columns) # type: ignore
        anchors = self.anchors(self.values) # type: ignore
        on_conflict = self.parse_conflict_handling()
        stmt = WRITE.format(table_name=ttable_name, col_names=columns, anchors=anchors, on_conflict=on_conflict)
        return (stmt, self.values) # type: ignore

    def write_many(self) -> tuple[str, list[list[Any]]]:

        # add validation for columns names ? and table_name ?

        if not all([self.table_name, self.target_columns, self.values]):
            raise ValueError()

        target_table = self.tables.get(self.table_name, None) # type: ignore
        if target_table is None:
            raise KeyError(f"Unknown table type: {self.table_name}")
        ttable_name = target_table._name
        columns = " ,".join(self.target_columns) # type: ignore
        anchors = self.anchors(self.values[0]) # type: ignore
        on_conflict = self.parse_conflict_handling()
        stmt =  WRITE.format(table_name=ttable_name, col_names=columns, anchors=anchors, on_conflict=on_conflict)
        return (stmt, self.values) # type: ignore

    def read_one(self) -> tuple[str, list[Any]]:
        node_table = self.tables["nodes"]._name
        fields = self.parse_fields()
        conditions, values = self.parse_conditions()
        joins = self.parse_joins()
        order_by = self.parse_order()
        axis = self.parse_axis()
        stmt = READ_ONE.format(
            fields=fields,
            node_table=node_table,
            joins=joins,
            conditions=conditions,
            order=order_by,
            axis=axis
        )
        return (stmt, values)

    def read_many(self) -> tuple[str, list[Any]]:
        node_table = self.tables["nodes"]._name
        fields = self.parse_fields()
        conditions, values = self.parse_conditions()
        joins = self.parse_joins()
        order_by = self.parse_order()
        axis = self.parse_axis()
        limit = self.parse_limit()
        stmt = READ_MANY.format(
            fields=fields,
            node_table=node_table,
            joins=joins,
            conditions=conditions,
            order=order_by,
            axis=axis,
            limit=limit
        )
        return (stmt, values)

    def delete(self) -> tuple[str, list[Any]]:
        node_table = self.tables["nodes"]._name
        conditions, values = self.parse_conditions(with_subqueries=True)
        stmt = DELETE.format(node_table=node_table, conditions=conditions)
        return (stmt, values)

    def update(self) -> tuple[str, list[Any]]:
        table = self.tables.get(self.table_name, None) # type: ignore
        if table is None:
            raise KeyError(f"Unknown table type: {self.table_name}")
        if self.setter is None:
            raise ValueError(f"You must Set some pairs of key values to update.")
        setter, svalues = self.update_setter(self.setter)
        conditions, cvalues = self.parse_conditions()
        stmt = UPDATE.format(table_name=table._name, setter=setter, conditions=conditions)
        return (stmt, svalues + cvalues)

    def parse_joins(self) -> str:
        def parse_sequence(sequence:list[str]|Conditions) -> list[str]:
            if sequence is None or len(sequence) == 0:
                return []
            if isinstance(sequence[0], Sequence) and not isinstance(sequence[0], str):
                fields = self.extract_fieldnames(sequence) # type: ignore
            else:
                fields = sequence

            tnodes = self.tables["nodes"]._name
            buff = []
            for fname in fields:
                if fname == "*":
                    continue
                namespace = self.namespaces.get(fname, None) # type: ignore
                if namespace is None:
                    raise KeyError(f"Unknown table field: {fname}")
                if namespace.is_joinable():
                    buff.append(namespace.join(tnodes))
            return buff

        tnodes = self.tables["nodes"]._name
        stmts = [self.namespaces["depth"].join(tnodes)]
        for sequence in [self.conds, self.order_by, self.fields]:
            stmts.extend(parse_sequence(sequence))
        return " ".join(list(set(stmts)))

    def parse_conditions(self, with_subqueries: bool=False) -> tuple[str, list[Any]]:
        """
        a set of conditions is a list of list condition tuple, possibly seperated by AND/OR operators.
        """
        if self.conds is None:
            return ("", [])

        segments, values = [], []
        for cond in self.conds:
            if isinstance(cond, Sequence) and not isinstance(cond, str):
                set_of_conds, vals = self.parse_set_of_conditions(cond, with_subqueries)
                segments.append(set_of_conds)
                values.extend(vals)
            elif isinstance(cond, str) and cond in ["AND", "OR"]:
                segments.append(cond)
            else:
                raise ValueError("Conditions must be a list of tuple[fieldName, op, value] | AND | OR")
        conditions = self.join_conditions(segments, comma=True)
        return (f"WHERE {conditions}", values)

    def parse_set_of_conditions(self, conds: list[tuple[str, str, Any] | str], with_subqueries: bool=False) -> tuple[str, list[Any]]:
        conditions, values = [], []
        for cond in conds:
            if with_subqueries and isinstance(cond, Sequence) and not isinstance(cond, str):
                converter = SqlConverter(
                    namespaces=self.namespaces,
                    tables=self.tables,
                    fields=["id"],
                    conds=[[cond]]
                )
                subquery, vals = converter.read_many()
                stmt = SEARCH_SUBQUERY.format(subquery=subquery.rstrip(";"))
                conditions.append(stmt)
                values.extend(vals)

            elif isinstance(cond, Sequence) and not isinstance(cond, str):
                f, op, val = cond
                namespace = self.namespaces.get(f, None)
                if namespace is None:
                    raise KeyError(f"Unknown Table Field: {f}")
                if op.upper() not in SqlOperator.values():
                    raise KeyError(f"Unknown Operator: {op}")

                where, val = namespace.where(op, val)
                conditions.append(where)
                if isinstance(val, list):
                    values.extend(val)
                else:
                    values.append(val)

            elif isinstance(cond, str) and cond in ["AND", "OR"]:
                conditions.append(cond)
            else:
                raise ValueError(f"Unknown Condition: {cond}")
        return (self.join_conditions(conditions), values)

    def join_conditions(self, segments: list[str], comma: bool = False) -> str:
        buff = []
        for i in range(len(segments)):
            seg = segments[i]
            if i % 2 != 0 and seg not in ["AND", "OR"]:
                buff.append("AND")
            if comma and seg not in ["AND", "OR"]:
                buff.append(f"({seg})")
            else:
                buff.append(seg)
        return " ".join(buff)

    def extract_fieldnames(self, sequence: Conditions) -> list[str]:
        buff = []
        for cond_set in sequence:
            if isinstance(cond_set, Sequence) and not isinstance(cond_set, str):
                for cond in cond_set:
                    if isinstance(cond, Sequence) and not isinstance(cond, str):
                        buff.append(cond[0])
        return buff

    def parse_fields(self) -> str:
        fields = ["*"]
        if self.fields is not None:
            fields = []
            for fname in self.fields:
                namespace = self.namespaces.get(fname, None)
                if namespace is None:
                    raise KeyError(f"Unknown field name: {fname}")
                fields.append(namespace.select())
        return ", ".join(fields)

    def parse_conflict_handling(self) -> str:
        handler = ""
        if self.on_conflict is not None:
            handler = self.on_conflict.sql()
        return handler

    def parse_limit(self) -> str:
        limit = ""
        if self.limit is not None:
            limit = f"LIMIT {str(self.limit)}"
        return limit

    def parse_order(self) -> str:
        if self.order_by is None:
            return ""
        f = ", ".join([self.namespaces[fname].select() for fname in self.order_by])
        return f"ORDER BY {f}"

    def parse_axis(self) -> str:
        if self.order_by is None:
            return ""
        match self.axis:
            case 1:
                return "ASC"
            case 0:
                return "DESC"
            case _:
                raise ValueError("Axis must be either 1 or 0")

    def update_setter(self, set_values: list[tuple[str, Any]]) -> tuple[str, Any]:
        values, fields = [], []
        for fname, val in set_values:
            namespace = self.namespaces.get(fname, None)
            if namespace is None:
                raise KeyError(f"Unknown field name: {fname}")
            fields.append(f"{fname} = ?")
            values.append(val)
        return (", ".join(fields), values)

    @staticmethod
    def condition_anchor(op: str, values: Any) -> str:
        """define the right anchor for the given condition operator."""
        anchor = "?"
        if op.lower() == "in" and isinstance(values, list):
            anchors = ' ,'.join(["?" for _ in range(len(values))])
            anchor = f"({anchors})"
        return anchor

    @staticmethod
    def anchors(values: list[Any]) -> str:
        return ' ,'.join(["?" for _ in range(len(values))])





from __future__ import annotations

import copy
import json
from pathlib import Path
from hashlib import sha1
from collections import deque
from itertools import chain
from typing import Literal, Optional, Any

from weetags.engine.engine import TreeEngine
from weetags.utils import valid_creation, valid_update, valid_append, apply_handler, ErrorHandler

Nid = str
StrOrPath = str | Path
Conditions = list[list[tuple[str, str, Any] | str] | str] | None
Setter = list[tuple[str, Any]]
Fields = list[str] | None
Node = dict[str, Any] | None
Nodes = list[dict[str, Any]]
Style = Literal["ascii", "ascii-ex", "ascii-exr", "ascii-emh", "ascii-emv", "ascii-em"]
Relations = Literal["parent", "children", "siblings", "ancestors", "descendants"]



# @apply_handler(ErrorHandler)
class Tree(TreeEngine):
    """
    A Tree Reprensation based on the Sqlite Engine. Able to realise basic graph operations on trees.
    :attributes:
        :name: (str). name of the tree
        :tables: (dict[str, Table]). Sqlites tables schema storing the tree.
        :namespace: (dict[FieldName, Namespace]) namespace representation of the tree data.
        :root_id: (str) id of the root node.
        :tree_size: (int) number of nodes contained in the tree.
        :tree_depth: (int) maximum number of depth in the tree.
        :info: (dict[str, Any]) summary of tree data.
    :warnings:
        :efficiency: As SQlite is not a native Graphdb, Large operation recquiring to walk accross the whole tree tend to be inneficients.
        Large but relatively light trees can be better off Being cached rather than stored in a database.
    """

    def __init__(
        self,
        tree_name: str,
        database: Optional[str] = ":memory:",
        timeout: float = 5,
        **params: Any) -> None:
        super().__init__(tree_name, database, timeout, **params)
        self._build_tree_context(tree_name)
        self.name = tree_name
        self.remove_orphans = True
        self.root_id = None
        if self.tree_size > 0:
            self.root_id = self.root

    def __repr__(self) -> str:
        return f"<Tree name: {self.name}, size: {self.tree_size}, depth: {self.tree_depth}>"

    @property
    def tree_size(self) -> int:
        nodes = self.tables["nodes"]._name
        return self._table_size(nodes)

    @property
    def tree_depth(self) -> int:
        metadata = self.tables["metadata"]._name
        return self._max_depth(metadata)

    @property
    def root(self) -> dict[str, Any]:
        return self._read_one(fields=["id"], conditions=[[("depth", "=", 0)]])["id"]

    @property
    def info(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "uri": self.uri,
            "size": self.tree_size,
            "depth": self.tree_depth,
            "model": {f.fname:f.ftype for f in self.namespaces.values()}
        }

    def export(self, path: StrOrPath, fields: Optional[Fields] = None) -> None:
        filtered = ["depth", "is_root","is_leaf", "nid"]
        base = self.node(self.root_id, fields)
        if base is None:
            raise KeyError("Root Node Not found")
        if fields is not None and any([bool(f) for f in fields if f in filtered]):
            raise KeyError(f'metadata Fields cannot be exported `["depth", "is_root","is_leaf", "nid"]`')

        with open(path, "w+") as f:
            node = {k:v for k,v in base.items() if k not in filtered}
            f.write(f"{json.dumps(base)}\n")
            for node in self.descendants_nodes(self.root_id, fields):
                node = {k:v for k,v in node.items() if k not in filtered}
                f.write(f"{json.dumps(node)}\n")

    def node(self, nid: Nid, fields: Fields = None) -> Node:
        return self._read_one(fields=fields, conditions=[[("id", "=", nid)]])

    def nodes_where(
        self,
        conditions: Optional[Conditions] = None,
        fields: Optional[Fields] = None,
        order_by: Optional[Fields] = None,
        axis: Optional[int] = 1,
        limit: Optional[int | None] = None
        ) -> Nodes:
        return self._read_many(fields, conditions, order_by, limit, axis)

    def nodes_relation_where(
        self,
        relation: Relations,
        conditions: Optional[Conditions] = None,
        fields: Optional[Fields] = None,
        order: Optional[Fields] = None,
        axis: Optional[int] = 1,
        limit: Optional[int | None] = None,
        include_base: bool = False,
    ) -> Nodes:
        relations = {
            "parent": self.parent_node, 
            "children": self.children_nodes, 
            "siblings": self.siblings_nodes, 
            "ancestors": self.ancestors_nodes, 
            "descendants": self.descendants_nodes
        }
        callback = relations.get(relation)
        nodes = self._nodes_where(conditions, list(set(["id"] + fields)), order, axis, limit)
        res = {}
        for node in nodes:
            related = callback(node["id"], fields)
            if include_base:
                res.update({sha1(json.dumps(node).encode()).hexdigest(): node})
            if isinstance(related, list):
                res.update({sha1(json.dumps(n).encode()).hexdigest():n for n in related})
            else:
                res.update({sha1(json.dumps(related).encode()).hexdigest(): related})
        return list(res.values())
            
    def parent_node(self, nid: Nid, fields: Optional[Fields] = None) -> Node:
        node = self.node(nid, ["id","parent"])
        if node is None:
            return None
        return self.node(node["parent"], fields)

    def children_nodes(
        self, 
        nid: Nid, 
        fields: Optional[Fields] = None,
        order_by: Optional[Fields] = None,
        axis: Optional[int] = 1,
        limit: Optional[int | None] = None
    ) -> Nodes:
        node = self.node(nid, ["id","children"])
        if node is None:
            return []
        return self.nodes_where([[("id","IN", node["children"])]], fields, order_by, axis, limit)

    def siblings_nodes(
        self, 
        nid: Nid, 
        fields: Optional[Fields] = None,
        order_by: Optional[Fields] = None,
        axis: Optional[int] = 1,
        limit: Optional[int | None] = None
        ) -> Nodes:
        node = self.node(nid, ["id","parent"])
        if node is None:
            return []
        pnode = self.node(node["parent"], ["children"])
        if pnode is None:
            return []
        return self.nodes_where([[("id", "IN", pnode["children"]), ("id", "!=", nid)]], fields, order_by, axis, limit)


    def ancestors_nodes(
        self,
        nid: Nid, 
        fields: Optional[Fields] = None,
        axis: Optional[int] = 1,
        limit: Optional[int | None] = None 
    ) -> Nodes:
        remove_parent= False
        if fields is not None and "parent" not in fields:
            # force parent, as needed to go up ancestors
            remove_parent = True
            fields =  ["parent"] + fields


        node = self.node(nid, ["id","parent"])
        if node is None:
            return []

        ancestors = []
        while node["parent"]:
            node = self.node(node["parent"], fields)
            if remove_parent:
                payload = copy.deepcopy(node)
                payload.pop("parent", None)
                ancestors.append(payload)
            else:
                ancestors.append(node)

        if axis != 1 or limit is not None:
            ancestors = self._parse_selection(ancestors, None, axis, limit)
        return ancestors

    def descendants_nodes(
        self,
        nid: Nid,
        fields: Optional[Fields] = None,
        axis: Optional[int] = 1,
        limit: Optional[int | None] = None
    ) -> Nodes:
        node = self.node(nid, ["id","children"])
        if node is None:
            return []

        descendants, queue = [], deque(node["children"])
        while len(queue) > 0:
            cid = queue.pop()
            node = self.node(cid, fields)
            children = self.node(cid, ["children"])
            queue.extendleft(children["children"]) # weird to do that, have to do 2 I/O operations
            descendants.append(node)

        if axis != 1 or limit is not None:
            descendants = self._parse_selection(descendants, None, axis, limit)
        return descendants

    def orphans_nodes(
        self,
        fields: Optional[Fields] = None,
        order: Optional[Fields] = None,
        axis: Optional[int] = 1,
        limit: Optional[int | None] = None
        ) -> Nodes:
        orphans = self.nodes_where([[("parent","is", None)]], fields, order, axis, limit)
        for i in range(len(orphans)):
            if orphans[i]["id"] == self.root_id:
                orphans.pop(i)
                break
        return orphans

    def is_related(self, nid0: Nid, nid1: Nid, check_siblings: bool=False) -> bool:
        if nid0 == nid1:
            return True

        desc = [i["id"] for i in self.descendants_nodes(nid0, fields=["id"])]
        if nid1 in desc:
            return True

        ancs = [i["id"] for i in self.ancestors_nodes(nid0, fields=["id"])]
        if nid1 in ancs:
            return True

        if check_siblings:
            sibs = [i["id"] for i in self.siblings_nodes(nid0, fields=["id"])]
            if nid1 in sibs:
                return True
        return False

    def path(self, nid: Nid, to: Nid, fields: Optional[Fields] = None) -> Nodes:
        if fields is None:
            fields = []

        from_node = [self.node(nid, list(set(["id", "parent"] + fields)))]
        to_node = [self.node(to, list(set(["id", "parent"] + fields)))]
        meetup = False
        while meetup is False:
            if ((from_node[-1]["parent"] == to_node[-1]["id"]) or
                (to_node[-1]["parent"] == from_node[-1]["id"])):
                to_node.append(self.node(to_node[-1]["parent"], list(set(["id","parent"] + fields))))
                break

            if from_node[-1]["parent"] is not None:
                from_node.append(self.node(from_node[-1]["parent"], list(set(["id","parent"] + fields))))
            if to_node[-1]["parent"] is not None:
                to_node.append(self.node(to_node[-1]["parent"], list(set(["id","parent"] + fields))))

            if from_node[-1]["id"] == to_node[-1]["id"]:
                meetup = True
        return from_node[:-1] + to_node[::-1]

    @valid_creation
    def add_node(self, node: Node) -> None:
        pid = node.get("parent", None)
        if pid is None and self.root_id is not None:
            raise ValueError("tree can only have one root")
        elif pid is None:
            self._add_node(node, 0, True, True)
        else:
            pnode = self._add_children(pid, node["id"])
            self._add_node(node, pnode["depth"] + 1)

    @valid_update
    def update_node(self, nid: Nid, set_values: Setter) -> None:
        self._update("nodes", set_values, [[("id", "=", nid)]])

    @valid_update
    def update_nodes_where(self, conditions: Conditions, set_values: Setter) -> None:
        nids = [n["id"] for n in self.nodes_where(conditions, ["id"])]
        self._update("nodes", set_values, [[("id", "IN", nids)]])

    @valid_append
    def append_node(self, nid: Nid, field_name: str, value: Any) -> None:
        base_value = self.node(nid, fields=[field_name]).get(field_name)
        self._update("nodes", [(field_name, base_value + [value])], [[("id","=",nid)]])

    @valid_append
    def extend_node(self, nid: Nid, field_name: str, values: list[Any]) -> None:
        base_value = self.node(nid, fields=[field_name]).get(field_name)
        self._update("nodes", [(field_name, base_value + values)], [[("id","=",nid)]])

    def delete_node(self, nid: Nid) -> None:
        self._delete_node(nid)
        if self.remove_orphans:
            self.delete_dead_branches()

    def delete_nodes_where(self, conditions: Optional[Conditions] = None) -> None:
        nodes = self.nodes_where(conditions, ["id"])
        for n in nodes:
            nid = n["id"]
            if nid == self.root_id:
                raise ValueError("cannot delete root node")
            self._delete_node(nid)
        if self.remove_orphans:
            self.delete_dead_branches()

    def delete_dead_branches(self) -> None:
        orphans = self.orphans_nodes(["id"])
        nodes = chain.from_iterable([[o] + self.descendants_nodes(o["id"], ["id"]) for o in orphans])
        [self._delete([[("id","=", node["id"])]]) for node in nodes]

    def delete_orphans(self):
        orphans = self.orphans_nodes(["id"])
        [self._delete([[("id","=", o["id"])]]) for o in orphans]

    def draw_tree(
        self,
        nid: Optional[Nid | None]  = None, 
        style: Optional[Style] = "ascii-ex", 
        extra_space: bool = False
    ) -> str:
        dt = {
            "ascii": ("|", "|-- ", "+-- "),
            "ascii-ex": ("\u2502", "\u251c\u2500\u2500 ", "\u2514\u2500\u2500 "),
            "ascii-exr": ("\u2502", "\u251c\u2500\u2500 ", "\u2570\u2500\u2500 "),
            "ascii-em": ("\u2551", "\u2560\u2550\u2550 ", "\u255a\u2550\u2550 "),
            "ascii-emv": ("\u2551", "\u255f\u2500\u2500 ", "\u2559\u2500\u2500 "),
            "ascii-emh": ("\u2502", "\u255e\u2550\u2550 ", "\u2558\u2550\u2550 "),
        }[style]

        if nid is None:
            nid = self.root_id

        root = self.node(nid, ["id", "parent", "children", "depth", "is_leaf"])
        if root["is_leaf"]:
            tree = f"{dt[2]}{root['id']}"
            return tree

        tree = f"{root['id']}\n"

        INITIAL_DEPTH = root["depth"]
        MAX_DEPTH = self.tree_depth
        BLOCK_SIZE = 2
        INDENTATION = 2
        LINED_SPACE = dt[0] + (" " * BLOCK_SIZE)
        EMPTY_SPACE = " " * (BLOCK_SIZE + 1)
        layer_state = [False] * (MAX_DEPTH - INITIAL_DEPTH)
        layer_state[0] =  bool(len(root["children"]))

        def _spacing(layer:int, layer_state: list[bool]):
            base_indentation = " " * INDENTATION
            layers = "".join([LINED_SPACE if v else EMPTY_SPACE for v in layer_state[:layer]])
            return base_indentation + layers

        def _draw(tree: str, queue: deque, layer_state: list[bool]):
            seen = set()
            while len(queue) > 0:
                nid = queue.popleft()
                node = self.node(nid, ["id", "parent", "children", "depth", "is_leaf"])
                layer = node["depth"] - INITIAL_DEPTH - 1

                if nid not in seen and len(node["children"]) == 0:
                    seen.add(nid)
                    queue.append(nid)
                    continue

                if len(queue) > 0:
                    space = _spacing(layer, layer_state)
                    tree += f"{space}{dt[1]}{node['id']}\n"

                else:
                    space = _spacing(layer, layer_state)
                    tree += f"{space}{dt[2]}{node['id']}\n"

                layer_state[(node["depth"] - INITIAL_DEPTH - 1)] = bool(len(queue))

                if extra_space and len(queue) == 0 and any(layer_state[:layer]) and len(node["children"])== 0:
                    space = _spacing(layer, layer_state)
                    tree += f"{space}\n"

                tree = _draw(tree, deque(node["children"]), layer_state)
            return tree
        return _draw(tree, deque(root["children"]), layer_state)

    def show_tree(
        self, 
        nid: Optional[Nid | None] = None, 
        style: Optional[Style] = "ascii-ex", 
        extra_space: bool=False
    ) -> None:
        tree = self.draw_tree(nid, style, extra_space)
        print(tree)

    def _add_node(self, node: Node, depth: int=0, is_root: bool= False, is_leaf: bool = True) -> None:
        self._write_one("nodes", list(node.keys()), list(node.values()), "none")
        self._write_one("metadata", ["nid", "depth", "is_root", "is_leaf"], [node["id"], depth, is_root, is_leaf], "none")

    def _add_children(self, nid: Nid, cnid: Nid) -> Node:
        pnode = self.node(nid, ["id", "depth", "children"])

        assert(pnode is not None)

        pnode.update({"children": list((set(pnode["children"] + [cnid])))})
        self._update("nodes", [("children", pnode["children"])], [[("id", "=", nid)]])
        self._update("metadata", [("is_leaf", False)], [[("nid","=", pnode["id"])]])
        return pnode

    def _delete_node(self, nid: Nid) -> None:
        if nid == self.root_id:
            raise ValueError("cannot delete root node")

        node = self.node(nid, ["children","parent"])
        [self._remove_parent(c) for c in node["children"]]
        self._remove_children(node["parent"], nid)
        self._delete([[("id", "=", nid)]])

    def _remove_children(self, nid: Nid, cnid: Nid):
        node = self.node(nid, ["id", "children"])
        if cnid in node["children"]:
            node["children"].remove(cnid)
        self._update("nodes", [("children", node["children"])], [[("id", "=", node["id"])]])

    def _remove_parent(self, nid: Nid):
        node = self.node(nid, ["id"])
        self._update("nodes", [("parent", None)], [[("id", "=", node["id"])]])

    def _parse_selection(
        self, 
        nodes: Nodes,
        order_by: Fields,
        axis: int,
        limit: int | None
    ) -> list[dict[str,Any]]:
        if order_by:
            nodes = sorted(nodes, key= lambda x: [x[f] for f in order_by])
        if axis == 1 and limit:
            nodes = nodes[:limit]
        elif axis == 0 and limit:
            nodes = nodes[::-1][:limit]
        elif axis == 0 and limit is None:
            nodes = nodes[::-1]
        return nodes